# vim: set ft=python ts=4 sw=4 expandtab:

from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from apologies.game import GameMode
from apologies.rules import Rules

from apologiesserver.event import EventHandler, RequestContext, TaskQueue
from apologiesserver.interface import (
    ActivityState,
    AdvertiseGameContext,
    AvailableGamesContext,
    CancelledReason,
    ConnectionState,
    ExecuteMoveContext,
    GameAdvertisedContext,
    GameCancelledContext,
    GameCompletedContext,
    GameIdleContext,
    GameInactiveContext,
    GameInvitationContext,
    GameJoinedContext,
    GamePlayerChangeContext,
    GamePlayerQuitContext,
    GamePlayerTurnContext,
    GameStartedContext,
    GameStateChangeContext,
    JoinGameContext,
    Message,
    MessageType,
    PlayerIdleContext,
    PlayerInactiveContext,
    PlayerMessageReceivedContext,
    PlayerRegisteredContext,
    PlayerState,
    PlayerType,
    PlayerUnregisteredContext,
    ProcessingError,
    RegisteredPlayersContext,
    RegisterPlayerContext,
    ReregisterPlayerContext,
    SendMessageContext,
    Visibility,
)
from tests.conftest import coroutine_mock, to_date


class TestTaskQueue:
    """
    Test the TaskQueue class.
    """

    def test_disconnect(self):
        socket1 = MagicMock()
        socket2 = MagicMock()

        queue = TaskQueue()
        assert queue.is_empty() is True
        assert queue.disconnects == set()

        queue.disconnect(None)
        assert queue.is_empty() is True
        assert queue.disconnects == set()

        queue.disconnect(socket1)
        assert queue.is_empty() is False
        assert queue.disconnects == {socket1}

        queue.disconnect(socket2)
        assert queue.is_empty() is False
        assert queue.disconnects == {socket1, socket2}

        queue.disconnect(socket1)
        assert queue.is_empty() is False
        assert queue.disconnects == {socket1, socket2}

        queue.clear()
        assert queue.is_empty() is True
        assert queue.disconnects == set()

    def test_message(self):
        socket1 = MagicMock()
        socket2 = MagicMock()
        socket3 = MagicMock()

        message1 = MagicMock()
        message1.to_json.return_value = "json1"

        message2 = MagicMock()
        message2.to_json.return_value = "json2"

        message3 = MagicMock()
        message3.to_json.return_value = "json3"

        player = MagicMock(websocket=socket2)

        queue = TaskQueue()
        assert queue.is_empty() is True
        assert queue.messages == []

        queue.message(message1, websockets=None, players=None)
        assert queue.is_empty() is True
        assert queue.messages == []

        queue.message(message1, websockets=[], players=[])
        assert queue.is_empty() is True
        assert queue.messages == []

        queue.message(message1, websockets=[socket1])
        assert queue.is_empty() is False
        assert queue.messages == [("json1", socket1)]

        queue.message(message2, players=[player])
        assert queue.is_empty() is False
        assert queue.messages == [("json1", socket1), ("json2", socket2)]

        queue.message(message3, websockets=[socket1, socket3], players=[player])
        assert queue.is_empty() is False
        assert queue.messages == [
            ("json1", socket1),
            ("json2", socket2),
            ("json3", socket1),
            ("json3", socket3),
            ("json3", socket2),
        ]

        queue.clear()
        assert queue.is_empty() is True
        assert queue.messages == []

    @pytest.mark.asyncio
    @patch("apologiesserver.event.asyncio")
    async def test_execute_empty(self, stub):
        stub.wait = coroutine_mock()
        queue = TaskQueue()
        await queue.execute()
        stub.wait.assert_not_awaited()  # wait() doesn't accept an empty list, so we just don't call it

    @pytest.mark.asyncio
    @patch("apologiesserver.event.asyncio")
    async def test_execute(self, stub):
        stub.wait = coroutine_mock()

        socket1 = MagicMock()
        socket2 = MagicMock()
        socket3 = MagicMock()

        message1 = MagicMock()
        message1.to_json.return_value = "json1"

        message2 = MagicMock()
        message2.to_json.return_value = "json2"

        queue = TaskQueue()
        queue.disconnect(socket1)
        queue.message(message1, websockets=[socket1])  # will be ignored because socket is disconnected
        queue.message(message2, websockets=[socket2, socket3])
        await queue.execute()

        # unfortunately, I can't find a good way to validate the tasks passed to this
        stub.wait.assert_awaited()


class TestEventHandler:
    """
    Test the basic EventHandler functionality.
    """

    def test_context_manager(self):
        manager = MagicMock()
        queue = MagicMock()
        queue.clear = MagicMock()
        handler = EventHandler(manager, queue)
        with handler as result:
            assert handler is result  # proves that the right object is returned
            queue.clear.assert_called_once()  # proves that we clear the queue on entering
            queue.clear = MagicMock()  # clear it so we can easily check for 2nd call
        queue.clear.assert_called_once()  # proves that we clear the cache on exiting

    @pytest.mark.asyncio
    async def test_execute(self):
        manager = MagicMock()
        queue = AsyncMock()
        queue.execute = coroutine_mock()
        handler = EventHandler(manager, queue)
        await handler.execute_tasks()
        queue.execute.assert_awaited_once()


class TestTaskMethods:
    """
    Test the task-related methods on EventHandler.
    """

    @patch("apologiesserver.event.arrow_utcnow")
    @patch("apologiesserver.event.config")
    def test_handle_idle_websocket_check_task(self, config, now):
        config.return_value = MagicMock(websocket_idle_thresh_min=10, websocket_inactive_thresh_min=20)
        now.return_value = to_date("2020-05-11T10:22:00,000")

        w1 = MagicMock()
        w2 = MagicMock()
        w3 = MagicMock()
        w4 = MagicMock()
        w5 = MagicMock()
        w6 = MagicMock()
        w7 = MagicMock()
        w8 = MagicMock()

        # Results: 1 is active (a1), 2 are idle (a2, a3) and 1 is inactive (a4)
        a1 = (w1, to_date("2020-05-11T10:12:00,001"), 0)  # active (9:59.999), no registered players
        a2 = (w2, to_date("2020-05-11T10:12:00,000"), 0)  # idle (10:00.000), no registered players
        a3 = (w3, to_date("2020-05-11T10:02:00,001"), 0)  # idle (19:59.999), no registered players
        a4 = (w4, to_date("2020-05-11T10:02:00,000"), 0)  # inactive (20:00.000), no registered players
        a5 = (w5, to_date("2020-05-11T10:12:00,001"), 5)  # active (9:59.999), with registered players
        a6 = (w6, to_date("2020-05-11T10:12:00,000"), 1)  # idle (10:00.000), with registered players
        a7 = (w7, to_date("2020-05-11T10:02:00,001"), 2)  # idle (19:59.999), with registered players
        a8 = (w8, to_date("2020-05-11T10:02:00,000"), 20)  # inactive (20:00.000), with registered players
        activity = [
            a1,
            a2,
            a3,
            a4,
            a5,
            a6,
            a7,
            a8,
        ]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_websocket_activity.return_value = activity
        handler.handle_websocket_idle_event = MagicMock()
        handler.handle_websocket_inactive_event = MagicMock()

        assert handler.handle_idle_websocket_check_task() == (2, 1)  # anything with a registered player is ignored

        idle_calls = [call(w2), call(w3)]
        inactive_calls = [call(w4)]

        handler.handle_websocket_idle_event.assert_has_calls(idle_calls)
        handler.handle_websocket_inactive_event.assert_has_calls(inactive_calls)

    @patch("apologiesserver.event.arrow_utcnow")
    @patch("apologiesserver.event.config")
    def test_handle_idle_player_check_task(self, config, now):
        config.return_value = MagicMock(player_idle_thresh_min=10, player_inactive_thresh_min=20)
        now.return_value = to_date("2020-05-11T10:22:00,000")

        p1 = MagicMock()
        p2 = MagicMock()
        p3 = MagicMock()
        p4 = MagicMock()
        p5 = MagicMock()
        p6 = MagicMock()
        p7 = MagicMock()
        p8 = MagicMock()

        # Results: 2 are active (a1, a5), 2 are idle (a2, a3) and 4 are inactive (a4, a6, a7, a8)
        # A disconnected player is either active or inactive, it can't be idle
        a1 = (p1, to_date("2020-05-11T10:12:00,001"), ConnectionState.CONNECTED)  # connected and active (9:59.999)
        a2 = (p2, to_date("2020-05-11T10:12:00,000"), ConnectionState.CONNECTED)  # connected and idle (10:00.000)
        a3 = (p3, to_date("2020-05-11T10:02:00,001"), ConnectionState.CONNECTED)  # connected and idle (19:59.999)
        a4 = (p4, to_date("2020-05-11T10:02:00,000"), ConnectionState.CONNECTED)  # connected and inactive (20:00.000)
        a5 = (p5, to_date("2020-05-11T10:12:00,001"), ConnectionState.DISCONNECTED)  # disconnected and active (9:59.999)
        a6 = (p6, to_date("2020-05-11T10:12:00,000"), ConnectionState.DISCONNECTED)  # disconnected and idle (10:00.000)
        a7 = (p7, to_date("2020-05-11T10:12:00,000"), ConnectionState.DISCONNECTED)  # disconnected and idle (19:59.999)
        a8 = (p8, to_date("2020-05-11T10:02:00,000"), ConnectionState.DISCONNECTED)  # disconnected and inactive (20:00.000)
        activity = [a1, a2, a3, a4, a5, a6, a7, a8]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_player_activity.return_value = activity
        handler.handle_player_idle_event = MagicMock()
        handler.handle_player_inactive_event = MagicMock()

        assert handler.handle_idle_player_check_task() == (2, 4)

        idle_calls = [call(p2), call(p3)]
        inactive_calls = [call(p4), call(p6), call(p7), call(p8)]

        handler.handle_player_idle_event.assert_has_calls(idle_calls)
        handler.handle_player_inactive_event.assert_has_calls(inactive_calls)

    @patch("apologiesserver.event.arrow_utcnow")
    @patch("apologiesserver.event.config")
    def test_handle_idle_game_check_task(self, config, now):
        config.return_value = MagicMock(game_idle_thresh_min=10, game_inactive_thresh_min=20)
        now.return_value = to_date("2020-05-11T10:22:00,000")

        g1 = MagicMock()
        g2 = MagicMock()
        g3 = MagicMock()
        g4 = MagicMock()

        # Results: 1 is active (a1), 2 are idle (a2, a3) and 1 is inactive (a4)
        a1 = (g1, to_date("2020-05-11T10:12:00,001"))  # active (9:59.999)
        a2 = (g2, to_date("2020-05-11T10:12:00,000"))  # idle (10:00.000)
        a3 = (g3, to_date("2020-05-11T10:02:00,001"))  # idle (19:59.999)
        a4 = (g4, to_date("2020-05-11T10:02:00,000"))  # inactive (20:00.000)
        activity = [a1, a2, a3, a4]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_game_activity.return_value = activity
        handler.handle_game_idle_event = MagicMock()
        handler.handle_game_inactive_event = MagicMock()

        assert handler.handle_idle_game_check_task() == (2, 1)

        idle_calls = [call(g2), call(g3)]
        inactive_calls = [call(g4)]

        handler.handle_game_idle_event.assert_has_calls(idle_calls)
        handler.handle_game_inactive_event.assert_has_calls(inactive_calls)

    @patch("apologiesserver.event.arrow_utcnow")
    @patch("apologiesserver.event.config")
    def test_handle_obsolete_game_check_task(self, config, now):
        config.return_value = MagicMock(game_retention_thresh_min=10)
        now.return_value = to_date("2020-05-11T10:22:00,000")

        g1 = MagicMock()
        g2 = MagicMock()

        # Results: 1 is young enough to keep (a1) and 1 is obsolete (a2)
        a1 = (g1, to_date("2020-05-11T10:12:00,001"))  # young enough to keep (9:59.999)
        a2 = (g2, to_date("2020-05-11T10:12:00,000"))  # obsolete (10:00.000)
        completion = [a1, a2]

        handler = EventHandler(MagicMock())
        handler.manager.lookup_game_completion.return_value = completion
        handler.handle_game_obsolete_event = MagicMock()

        assert handler.handle_obsolete_game_check_task() == 1

        obsolete_calls = [call(g2)]

        handler.handle_game_obsolete_event.assert_has_calls(obsolete_calls)


class TestRequestMethods:
    """
    Test the request-related methods on EventHandler.
    """

    @patch("apologiesserver.event.config")
    def test_handle_register_player_request_below_limit(self, config):
        config.return_value = MagicMock(registered_player_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_registered_player_count.return_value = 4
        handler.handle_player_registered_event = MagicMock()
        websocket = MagicMock()
        context = RegisterPlayerContext(handle="leela")
        message = Message(MessageType.REGISTER_PLAYER, context=context)
        handler.handle_register_player_request(message, websocket)
        assert handler.queue.is_empty()
        handler.handle_player_registered_event.assert_called_once_with(websocket, "leela")

    @patch("apologiesserver.event.config")
    def test_handle_register_player_request_at_limit(self, config):
        config.return_value = MagicMock(registered_player_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_registered_player_count.return_value = 5
        handler.handle_player_registered_event = MagicMock()
        websocket = MagicMock()
        context = RegisterPlayerContext(handle="leela")
        message = Message(MessageType.REGISTER_PLAYER, context=context)
        with pytest.raises(ProcessingError, match=r"System user limit reached"):
            handler.handle_register_player_request(message, websocket)
        assert handler.queue.is_empty()
        handler.handle_player_registered_event.assert_not_called()

    def test_handle_reregister_player_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_player_reregistered_event = MagicMock()
        context = ReregisterPlayerContext(handle="leela")
        message = Message(MessageType.REREGISTER_PLAYER, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_reregister_player_request(request)
        handler.handle_player_reregistered_event.assert_called_once_with(player, websocket)

    def test_handle_unregister_player_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_player_unregistered_event = MagicMock()
        message = Message(MessageType.UNREGISTER_PLAYER, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_unregister_player_request(request)
        handler.handle_player_unregistered_event.assert_called_once_with(player, game)

    def test_handle_list_players_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_registered_players_event = MagicMock()
        message = Message(MessageType.LIST_PLAYERS, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_list_players_request(request)
        handler.handle_registered_players_event.assert_called_once_with(player)

    def test_handle_advertise_game_request_already_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_advertised_event = MagicMock()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is already playing a game"):
            handler.handle_advertise_game_request(request)
        handler.handle_game_advertised_event.assert_not_called()

    @patch("apologiesserver.event.config")
    def test_handle_advertise_game_request_below_limit(self, config):
        config.return_value = MagicMock(total_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_total_game_count.return_value = 4
        handler.handle_game_advertised_event = MagicMock()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        handler.handle_advertise_game_request(request)
        handler.handle_game_advertised_event.assert_called_once_with(player, context)

    @patch("apologiesserver.event.config")
    def test_handle_advertise_game_request_at_limit(self, config):
        config.return_value = MagicMock(total_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_total_game_count.return_value = 5
        handler.handle_game_advertised_event = MagicMock()
        context = AdvertiseGameContext("name", GameMode.STANDARD, 3, Visibility.PUBLIC, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"System game limit reached"):
            handler.handle_advertise_game_request(request)
        handler.handle_game_advertised_event.assert_not_called()

    def test_handle_list_available_games_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_available_games_event = MagicMock()
        message = Message(MessageType.LIST_AVAILABLE_GAMES, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_list_available_games_request(request)
        handler.handle_available_games_event.assert_called_once_with(player)

    def test_handle_join_game_request_already_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_joined_event = MagicMock()
        context = JoinGameContext(game_id="game")
        message = Message(MessageType.JOIN_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is already playing a game"):
            handler.handle_join_game_request(request)
        handler.handle_game_joined_event.assert_not_called()

    def test_handle_join_game_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_joined_event = MagicMock()
        context = JoinGameContext(game_id="game")
        message = Message(MessageType.JOIN_GAME, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        handler.handle_join_game_request(request)
        handler.handle_game_joined_event.assert_called_once_with(player, game_id="game")

    def test_handle_quit_game_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_not_called()

    def test_handle_quit_game_request_not_in_progress(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_in_progress.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not in progress"):
            handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_not_called()

    def test_handle_quit_game_request_advertiser(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Advertiser may not quit a game"):
            handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_not_called()

    def test_handle_quit_game_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_quit_event = MagicMock()
        message = Message(MessageType.QUIT_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_quit_game_request(request)
        handler.handle_game_player_quit_event.assert_called_once_with(player, game)

    def test_handle_start_game_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    def test_handle_start_game_request_already_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = True
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is already being played"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    def test_handle_start_game_request_not_advertiser(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="bender")
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player did not advertise this game"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    @patch("apologiesserver.event.config")
    def test_handle_start_game_request_below_limit(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_in_progress_game_count.return_value = 4
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_called_once_with(game)

    @patch("apologiesserver.event.config")
    def test_handle_start_game_request_at_limit(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        handler = EventHandler(MagicMock())
        handler.manager.get_in_progress_game_count.return_value = 5
        handler.handle_game_started_event = MagicMock()
        message = Message(MessageType.START_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"System game limit reached"):
            handler.handle_start_game_request(request)
        handler.handle_game_started_event.assert_not_called()

    def test_handle_cancel_game_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_cancel_game_request_not_in_progress(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_in_progress.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not in progress"):
            handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_cancel_game_request_not_advertiser(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="bender")
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player did not advertise this game"):
            handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_cancel_game_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        request = RequestContext(message, websocket, player, game)
        handler.handle_cancel_game_request(request)
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.CANCELLED)

    def test_handle_execute_move_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_execute_move_request(request)
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request_not_being_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not being played"):
            handler.handle_execute_move_request(request)
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request_no_move_pending(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"No move is pending for this player"):
            handler.handle_execute_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.is_legal_move.assert_not_called()
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request_illegal_move(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = True
        game.is_legal_move.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"The chosen move is not legal"):
            handler.handle_execute_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.is_legal_move.assert_called_once_with("leela", "move")
        handler.handle_game_player_move_event.assert_not_called()

    def test_handle_execute_move_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_player_move_event = MagicMock()
        context = ExecuteMoveContext(move_id="move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = True
        request = RequestContext(message, websocket, player, game)
        handler.handle_execute_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        handler.handle_game_player_move_event.assert_called_once_with(player, game, "move")

    def test_handle_optimal_move_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_optimal_move_request(request)
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_optimal_move_request_not_being_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not being played"):
            handler.handle_optimal_move_request(request)
        game.mark_active.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_optimal_move_request_no_move_pending(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"No move is pending for this player"):
            handler.handle_optimal_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.mark_active.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_optimal_move_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        websocket = MagicMock()
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_move_pending.return_value = True
        request = RequestContext(message, websocket, player, game)
        handler.handle_optimal_move_request(request)
        game.is_move_pending.assert_called_once_with("leela")
        game.mark_active.assert_called_once()
        handler.handle_game_programmatic_move_event.assert_called_once_with("leela", game)

    def test_handle_retrieve_game_state_request_not_playing(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_state_change_event = MagicMock()
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = None
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Player is not playing a game"):
            handler.handle_retrieve_game_state_request(request)
        handler.handle_game_state_change_event.assert_not_called()

    def test_handle_retrieve_game_state_request_not_being_played(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_state_change_event = MagicMock()
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = False
        request = RequestContext(message, websocket, player, game)
        with pytest.raises(ProcessingError, match=r"Game is not being played"):
            handler.handle_retrieve_game_state_request(request)
        handler.handle_game_state_change_event.assert_not_called()

    def test_handle_retrieve_game_state_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_game_state_change_event = MagicMock()
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        websocket = MagicMock()
        player = MagicMock()
        game = MagicMock()
        game.is_playing.return_value = True
        request = RequestContext(message, websocket, player, game)
        handler.handle_retrieve_game_state_request(request)
        handler.handle_game_state_change_event.assert_called_once_with(game, player)

    def test_handle_send_message_request(self):
        handler = EventHandler(MagicMock())
        handler.handle_player_message_received_event = MagicMock()
        context = SendMessageContext(message="hello", recipient_handles=["fry", "bender"])
        message = Message(MessageType.SEND_MESSAGE, player_id="id", context=context)
        websocket = MagicMock()
        player = MagicMock(handle="handle")
        game = MagicMock()
        request = RequestContext(message, websocket, player, game)
        handler.handle_send_message_request(request)
        handler.handle_player_message_received_event.assert_called_once_with("handle", ["fry", "bender"], "hello")


class TestEventMethods:
    """
    Test the event-related methods on EventHandler.
    """

    def test_handle_server_shutdown_event(self):
        websocket = MagicMock()
        game = MagicMock()
        message = Message(MessageType.SERVER_SHUTDOWN)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_all_websockets.return_value = [websocket]
        handler.manager.lookup_in_progress_games.return_value = [game]
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_server_shutdown_event()
        handler.queue.message.assert_called_once_with(message, websockets=[websocket])
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.SHUTDOWN, notify=False)

    @patch("apologiesserver.event.config")
    def test_handle_websocket_connected_event_below_limit(self, config):
        config.return_value = MagicMock(websocket_limit=5)
        websocket = MagicMock()
        handler = EventHandler(MagicMock())
        handler.manager.track_websocket = MagicMock()
        handler.manager.get_websocket_count.return_value = 4
        handler.handle_websocket_connected_event(websocket)
        handler.manager.track_websocket.assert_called_once_with(websocket)

    @patch("apologiesserver.event.config")
    def test_handle_websocket_connected_event_at_limit(self, config):
        config.return_value = MagicMock(websocket_limit=5)
        websocket = MagicMock()
        handler = EventHandler(MagicMock())
        handler.manager.track_websocket = MagicMock()
        handler.manager.get_websocket_count.return_value = 5
        with pytest.raises(ProcessingError, match=r"Connection limit reached"):
            handler.handle_websocket_connected_event(websocket)
        handler.manager.track_websocket.assert_not_called()

    def test_handle_websocket_disconnected_event(self):
        player = MagicMock()
        websocket = MagicMock()
        handler = EventHandler(MagicMock())
        handler.handle_player_disconnected_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.manager.delete_websocket = MagicMock()
        handler.manager.lookup_players_for_websocket.return_value = [player]
        handler.handle_websocket_disconnected_event(websocket)
        handler.handle_player_disconnected_event.assert_called_once_with(player)
        handler.manager.delete_websocket.assert_called_once_with(websocket)

    def test_handle_websocket_idle_event_no_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=ActivityState.IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_websocket_idle_event(websocket)
        websocket.mark_idle.assert_not_called()
        handler.queue.message.assert_not_called()

    def test_handle_websocket_idle_event_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=None)
        message = Message(MessageType.WEBSOCKET_IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_websocket_idle_event(websocket)
        websocket.mark_idle.assert_called_once()
        handler.queue.message.assert_called_once_with(message, websockets=[websocket.websocket])

    def test_handle_websocket_inactive_event_no_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=ActivityState.INACTIVE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.handle_websocket_inactive_event(websocket)
        websocket.mark_inactive.assert_not_called()
        handler.queue.message.assert_not_called()
        handler.queue.disconnect.assert_not_called()

    def test_handle_websocket_inactive_event_change(self):
        websocket = MagicMock(websocket=MagicMock(), activity_state=None)
        message = Message(MessageType.WEBSOCKET_INACTIVE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.handle_websocket_inactive_event(websocket)
        websocket.mark_inactive.assert_called_once()
        handler.queue.message.assert_called_once_with(message, websockets=[websocket.websocket])
        handler.queue.disconnect.assert_called_once_with(websocket.websocket)

    def test_handle_registered_players_event(self):
        player = MagicMock()
        result = MagicMock()
        registered = MagicMock()
        result.to_registered_player.return_value = registered
        context = RegisteredPlayersContext(players=[registered])
        message = Message(MessageType.REGISTERED_PLAYERS, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_all_players.return_value = [result]
        handler.handle_registered_players_event(player)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_available_games_event(self):
        player = MagicMock()
        advertised = MagicMock()
        game = MagicMock()
        game.to_advertised_game.return_value = advertised
        context = AvailableGamesContext(games=[advertised])
        message = Message(MessageType.AVAILABLE_GAMES, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_available_games = MagicMock()
        handler.manager.lookup_available_games.return_value = [game]
        handler.handle_available_games_event(player)
        handler.manager.lookup_available_games.assert_called_once_with(player)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_player_registered_event(self):
        websocket = MagicMock()
        player = MagicMock(player_id="player_id", handle="handle")
        context = PlayerRegisteredContext(handle="handle")
        message = Message(MessageType.PLAYER_REGISTERED, player_id="player_id", context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.track_player = MagicMock()
        handler.manager.track_player.return_value = player
        handler.handle_player_registered_event(websocket, "leela")
        handler.manager.track_player.assert_called_once_with(websocket, "leela")
        handler.queue.message.assert_called_once_with(message, websockets=[websocket])

    def test_handle_player_reregistered_event(self):
        websocket = MagicMock()
        player = MagicMock(player_id="player_id", handle="handle")
        context = PlayerRegisteredContext(handle="handle")
        message = Message(MessageType.PLAYER_REGISTERED, player_id="player_id", context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.retrack_player = MagicMock()
        handler.handle_player_reregistered_event(player, websocket)
        handler.manager.retrack_player.assert_called_once_with(player, websocket)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_player_unregistered_event_no_game(self):
        player = MagicMock(player_id="player_id", handle="leela")
        context = PlayerUnregisteredContext(handle="leela")
        message = Message(MessageType.PLAYER_UNREGISTERED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.delete_player = MagicMock()
        handler.handle_game_player_left_event = MagicMock()
        handler.handle_player_unregistered_event(player)
        player.mark_quit.assert_called_once()
        handler.handle_game_player_left_event.assert_not_called()
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.delete_player.assert_called_once_with(player)

    def test_handle_player_unregistered_event_with_game(self):
        comment = "Player leela unregistered"
        player = MagicMock(player_id="player_id", handle="leela")
        context = PlayerUnregisteredContext(handle="leela")
        message = Message(MessageType.PLAYER_UNREGISTERED, context=context)
        game = MagicMock()
        game.is_viable.return_value = True
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.delete_player = MagicMock()
        handler.handle_game_player_left_event = MagicMock()
        handler.handle_player_unregistered_event(player, game)
        player.mark_quit.assert_called_once()
        handler.handle_game_player_left_event.assert_called_once_with(player, game, comment)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.delete_player.assert_called_once_with(player)

    def test_handle_player_disconnected_event_no_game(self):
        player = MagicMock()
        handler = EventHandler(MagicMock())
        handler.handle_game_player_left_event = MagicMock()
        handler.manager.lookup_game = MagicMock()
        handler.manager.lookup_game.return_value = None
        handler.handle_player_disconnected_event(player)
        handler.manager.lookup_game.assert_called_once_with(player=player)
        player.mark_disconnected.assert_called_once()
        handler.handle_game_player_left_event.assert_not_called()

    def test_handle_player_disconnected_event_with_game(self):
        comment = "Player leela disconnected"
        player = MagicMock(handle="leela")
        game = MagicMock()
        game.is_viable.return_value = True
        handler = EventHandler(MagicMock())
        handler.handle_game_player_left_event = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_player_disconnected_event(player)
        handler.manager.lookup_game.assert_called_once_with(player=player)
        player.mark_disconnected.assert_called_once()
        handler.handle_game_player_left_event.assert_called_once_with(player, game, comment)

    def test_handle_player_idle_event_no_change(self):
        player = MagicMock(activity_state=ActivityState.IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_player_idle_event(player)
        handler.queue.message.assert_not_called()
        player.mark_idle.assert_not_called()

    def test_handle_player_idle_event_change(self):
        player = MagicMock(handle="handle", activity_state=None)
        context = PlayerIdleContext(handle="handle")
        message = Message(MessageType.PLAYER_IDLE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_player_idle_event(player)
        handler.queue.message.assert_called_once_with(message, players=[player])
        player.mark_idle.assert_called_once()

    def test_handle_player_inactive_event_no_change(self):
        websocket = MagicMock()
        player = MagicMock(websocket=websocket, activity_state=ActivityState.INACTIVE)
        handler = EventHandler(MagicMock())
        handler.handle_player_unregistered_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.handle_player_inactive_event(player)
        player.mark_inactive.assert_not_called()
        handler.queue.message.assert_not_called()
        handler.queue.disconnect.assert_not_called()
        handler.handle_player_unregistered_event.assert_not_called()

    def test_handle_player_inactive_event_change(self):
        websocket = MagicMock()
        player = MagicMock(handle="handle", websocket=websocket, activity_state=None)
        game = MagicMock()
        context = PlayerInactiveContext(handle="handle")
        message = Message(MessageType.PLAYER_INACTIVE, context=context)
        handler = EventHandler(MagicMock())
        handler.handle_player_unregistered_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.queue.disconnect = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_player_inactive_event(player)
        player.mark_inactive.assert_called_once()
        handler.manager.lookup_game.assert_called_once_with(player=player)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.queue.disconnect.assert_not_called()  # we don't close now, only when the websocket itself goes inactive
        handler.handle_player_unregistered_event.assert_called_once_with(player, game)

    def test_handle_player_message_received_event(self):
        fry = MagicMock()
        bender = MagicMock()
        context = PlayerMessageReceivedContext("leela", ["fry", "bender"], "hello")
        message = Message(MessageType.PLAYER_MESSAGE_RECEIVED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_player = MagicMock()
        handler.manager.lookup_player.side_effect = [fry, bender]
        handler.handle_player_message_received_event("leela", ["fry", "bender"], "hello")
        handler.manager.lookup_player.assert_has_calls([call(handle="fry"), call(handle="bender")])
        handler.queue.message.assert_called_once_with(message, players=[fry, bender])

    def test_handle_game_advertised_event(self):
        player = MagicMock()
        advertised = MagicMock()
        game = MagicMock()
        result = MagicMock()
        game.to_advertised_game.return_value = result
        context = GameAdvertisedContext(game=result)
        message = Message(MessageType.GAME_ADVERTISED, context=context)
        handler = EventHandler(MagicMock())
        handler.manager.track_game = MagicMock(return_value=game)
        handler.handle_game_invitation_event = MagicMock()
        handler.handle_game_joined_event = MagicMock()
        handler.queue.message = MagicMock()
        handler.handle_game_advertised_event(player, advertised)
        handler.manager.track_game.assert_called_once_with(player, advertised)
        handler.handle_game_joined_event.assert_called_once_with(player, game=game)
        handler.handle_game_invitation_event.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_game_invitation_event_none_invited(self):
        game = MagicMock(invited_handles=[])
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_player = MagicMock()
        handler.handle_game_invitation_event(game)
        handler.manager.lookup_player.assert_not_called()
        handler.queue.message.assert_not_called()

    def test_handle_game_invitation_event_with_invited(self):
        fry = MagicMock()
        bender = MagicMock()
        game = MagicMock(invited_handles=["fry", "bender"])
        advertised = MagicMock()
        game.to_advertised_game.return_value = advertised
        context = GameInvitationContext(advertised)
        message = Message(MessageType.GAME_INVITATION, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_player = MagicMock()
        handler.manager.lookup_player.side_effect = [fry, bender]
        handler.handle_game_invitation_event(game)
        handler.manager.lookup_player.assert_has_calls([call(handle="fry"), call(handle="bender")])
        handler.queue.message.assert_called_once_with(message, players=[fry, bender])

    def test_handle_game_joined_event_bad_call(self):
        player = MagicMock()
        handler = EventHandler(MagicMock())
        with pytest.raises(ProcessingError, match=r"Invalid arguments"):
            handler.handle_game_joined_event(player, game_id=None, game=None)  # need to pass either id or game

    def test_handle_game_joined_event_not_found(self):
        player = MagicMock(handle="handle")
        handler = EventHandler(MagicMock())
        handler.manager.lookup_game = MagicMock(return_value=None)
        with pytest.raises(ProcessingError, match=r"Unknown or invalid game"):
            handler.handle_game_joined_event(player, game_id="game_id")
        handler.manager.lookup_game.assert_called_once_with(game_id="game_id")

    def test_handle_game_joined_event_not_available(self):
        player = MagicMock(handle="handle")
        game = MagicMock()
        game.is_available.return_value = False
        handler = EventHandler(MagicMock())
        handler.manager.lookup_game = MagicMock(return_value=game)
        with pytest.raises(ProcessingError, match=r"Unknown or invalid game"):
            handler.handle_game_joined_event(player, game_id="game_id")
        handler.manager.lookup_game.assert_called_once_with(game_id="game_id")
        game.is_available.assert_called_once_with("handle")

    def test_handle_game_joined_event_pending_by_id(self):
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = False
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game_id="game_id")
        handler.manager.lookup_game.assert_called_once_with(game_id="game_id")
        game.is_available.assert_called_once_with("handle")
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_not_called()

    def test_handle_game_joined_event_pending_for_game(self):
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = False
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game=game)
        handler.manager.lookup_game.assert_not_called()
        game.is_available.assert_not_called()
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_not_called()

    @patch("apologiesserver.event.config")
    def test_handle_game_joined_event_fully_joined(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = True
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.get_in_progress_game_count.return_value = 4
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game=game)
        handler.manager.lookup_game.assert_not_called()
        game.is_available.assert_not_called()
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_called_once_with(game)

    @patch("apologiesserver.event.config")
    def test_handle_game_joined_event_game_limit(self, config):
        config.return_value = MagicMock(in_progress_game_limit=5)
        player = MagicMock(handle="handle")
        game = MagicMock(game_id="id", mode=GameMode.STANDARD, advertiser_handle="advertiser")
        game.name = "name"  # MagicMock treats the name attribute specially
        game.is_available.return_value = True
        game.is_fully_joined.return_value = True
        context = GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.get_in_progress_game_count.return_value = 5
        handler.manager.lookup_game = MagicMock(return_value=game)
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_started_event = MagicMock()
        handler.handle_game_joined_event(player, game=game)
        handler.manager.lookup_game.assert_not_called()
        game.is_available.assert_not_called()
        game.mark_active.assert_called_once()
        player.mark_joined.assert_called_once_with(game)
        game.mark_joined.assert_called_once_with("handle")
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Player joined game")
        handler.handle_game_started_event.assert_not_called()  # because in-progress limit was reached

    def test_handle_game_started_event(self):
        player = MagicMock()
        game = MagicMock(game_id="game")
        context = GameStartedContext(game_id="game")
        message = Message(MessageType.GAME_STARTED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.handle_game_next_turn_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_started_event(game)
        game.mark_active.assert_called_once()
        game.mark_started.assert_called_once()
        player.mark_playing.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_change_event.assert_called_once_with(game, "Game started")
        handler.handle_game_state_change_event.assert_called_once_with(game)
        handler.handle_game_next_turn_event.assert_called_once_with(game)

    def test_handle_game_cancelled_event_notify(self):
        player = MagicMock()
        game = MagicMock(game_id="game")
        context = GameCancelledContext("game", CancelledReason.SHUTDOWN, "comment")
        message = Message(MessageType.GAME_CANCELLED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_cancelled_event(game, CancelledReason.SHUTDOWN, "comment", notify=True)
        player.mark_quit.assert_called_once()
        game.mark_cancelled.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_state_change_event.assert_called_once_with(game)

    def test_handle_game_cancelled_event_no_notify(self):
        player = MagicMock()
        game = MagicMock()
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_cancelled_event(game, CancelledReason.SHUTDOWN, "comment", notify=False)
        player.mark_quit.assert_called_once()
        game.mark_cancelled.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_not_called()  # notification disabled
        handler.handle_game_state_change_event.assert_not_called()  # notification disabled

    def test_handle_game_completed_event(self):
        player = MagicMock()
        game = MagicMock(game_id="game")
        context = GameCompletedContext("game", "winner", "comment")
        message = Message(MessageType.GAME_COMPLETED, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_completed_event(game, "winner", "comment")
        player.mark_quit.assert_called_once()
        game.mark_completed.assert_called_once()
        handler.manager.lookup_game_players.assert_called_once_with(game)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_state_change_event.assert_called_once_with(game)

    def test_handle_game_idle_event_no_change(self):
        game = MagicMock(activity_state=ActivityState.IDLE)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_idle_event(game)
        handler.queue.message.assert_not_called()

    def test_handle_game_idle_event_change(self):
        player = MagicMock()
        game = MagicMock(game_id="game", activity_state=None)
        context = GameIdleContext(game_id="game")
        message = Message(MessageType.GAME_IDLE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_idle_event(game)
        handler.queue.message.assert_called_once_with(message, players=[player])

    def test_handle_game_inactive_event(self):
        game = MagicMock(activity_state=ActivityState.INACTIVE)
        handler = EventHandler(MagicMock())
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_inactive_event(game)
        game.mark_inactive.assert_not_called()
        handler.handle_game_cancelled_event.assert_not_called()

    def test_handle_game_inactive_event_change(self):
        player = MagicMock()
        game = MagicMock(game_id="game", activity_state=None)
        context = GameInactiveContext(game_id="game")
        message = Message(MessageType.GAME_INACTIVE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.manager.lookup_game_players.return_value = [player]
        handler.handle_game_inactive_event(game)
        game.mark_inactive.assert_called_once()
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.INACTIVE)

    def test_handle_game_obsolete_event(self):
        game = MagicMock()
        handler = EventHandler(MagicMock())
        handler.manager.delete_game = MagicMock()
        handler.handle_game_obsolete_event(game)
        handler.manager.delete_game.assert_called_once_with(game)

    def test_handle_game_player_quit_event(self):
        comment = "Player leela quit"
        player = MagicMock(handle="leela")
        game = MagicMock(game_id="game")
        context = GamePlayerQuitContext(handle="leela", game_id="game")
        message = Message(MessageType.GAME_PLAYER_QUIT, context=context)
        game.is_viable.return_value = True
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_left_event = MagicMock()
        handler.handle_game_player_quit_event(player, game)
        game.mark_active.assert_called_once()
        player.mark_quit.assert_called_once()
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.handle_game_player_left_event.assert_called_once_with(player, game, comment)

    def test_handle_game_player_left_event_advertiser(self):
        player = MagicMock(handle="leela")
        game = MagicMock(advertiser_handle="leela")
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.CANCELLED, "comment")
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_not_viable(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = False
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_called_once_with(game, CancelledReason.NOT_VIABLE, "comment")
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_viable_not_playing(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = True
        game.is_playing.return_value = False
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_viable_playing(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = True
        game.is_playing.return_value = True
        game.is_move_pending.return_value = False
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_not_called()

    def test_handle_game_player_left_event_viable_pending(self):
        player = MagicMock(handle="bender")
        game = MagicMock(advertiser_handle="leela")
        game.is_viable.return_value = True
        game.is_playing.return_value = True
        game.is_move_pending.return_value = True
        handler = EventHandler(MagicMock())
        handler.handle_game_player_change_event = MagicMock()
        handler.handle_game_cancelled_event = MagicMock()
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_left_event(player, game, "comment")
        game.mark_quit.assert_called_once_with("bender")
        handler.handle_game_player_change_event.assert_called_once_with(game, "comment")
        handler.handle_game_cancelled_event.assert_not_called()
        handler.handle_game_programmatic_move_event.assert_called_once_with("bender", game)

    def test_handle_game_player_move_event(self):
        player = MagicMock(handle="handle")
        game = MagicMock()
        handler = EventHandler(MagicMock())
        handler.handle_game_move_event = MagicMock()
        handler.handle_game_player_move_event(player, game, "move")
        handler.handle_game_move_event.assert_called_once_with("handle", game, "move")

    @patch("apologiesserver.event.RewardV1InputSource")
    def test_handle_game_programmatic_move_event(self, obj):
        source = MagicMock()
        obj.return_value = source  # mock the constructor
        view = MagicMock()
        move = MagicMock(id="move")
        game = MagicMock(mode=GameMode.STANDARD)
        game.get_player_view.return_value = view
        game.get_legal_moves.return_value = [move]
        source.choose_move.return_value = move
        handler = EventHandler(MagicMock())
        handler.handle_game_move_event = MagicMock()
        handler.handle_game_programmatic_move_event("handle", game)
        game.get_player_view.assert_called_once_with("handle")
        game.get_legal_moves.assert_called_once_with("handle")
        source.choose_move.assert_called_once_with(GameMode.STANDARD, view, [move], Rules.evaluate_move)
        handler.handle_game_move_event.assert_called_once_with("handle", game, "move")

    def test_handle_game_move_event_completed(self):
        game = MagicMock()
        game.execute_move.return_value = (True, "winner", "comment")
        handler = EventHandler(MagicMock())
        handler.handle_game_completed_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.handle_game_next_turn_event = MagicMock()
        handler.handle_game_move_event("handle", game, "move_id")
        game.mark_active.assert_called_once()
        game.execute_move.assert_called_once_with("handle", "move_id")
        handler.handle_game_state_change_event.assert_called_once_with(game)
        handler.handle_game_completed_event.assert_called_once_with(game, "winner", "comment")
        handler.handle_game_next_turn_event.assert_not_called()

    def test_handle_game_player_event_not_completed(self):
        game = MagicMock()
        game.execute_move.return_value = (False, None, None)
        handler = EventHandler(MagicMock())
        handler.handle_game_completed_event = MagicMock()
        handler.handle_game_state_change_event = MagicMock()
        handler.handle_game_next_turn_event = MagicMock()
        handler.handle_game_move_event("handle", game, "move_id")
        game.mark_active.assert_called_once()
        game.execute_move.assert_called_once_with("handle", "move_id")
        handler.handle_game_state_change_event.assert_called_once_with(game)
        handler.handle_game_completed_event.assert_not_called()
        handler.handle_game_next_turn_event.assert_called_once_with(game)

    def test_handle_game_next_turn_event_programmatic(self):
        game = MagicMock()
        game.get_next_turn.return_value = ("handle", PlayerType.PROGRAMMATIC)
        handler = EventHandler(MagicMock())
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_turn_event = MagicMock()
        handler.handle_game_next_turn_event(game)
        handler.handle_game_programmatic_move_event.assert_called_once_with("handle", game)
        handler.handle_game_player_turn_event.assert_not_called()

    def test_handle_game_next_turn_event_human_not_found(self):
        game = MagicMock()
        game.get_next_turn.return_value = ("handle", PlayerType.HUMAN)
        game.is_playing.return_value = True
        handler = EventHandler(MagicMock())
        handler.manager.lookup_player = MagicMock(return_value=None)  # player is not found
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_turn_event = MagicMock()
        handler.handle_game_next_turn_event(game)
        handler.handle_game_programmatic_move_event.assert_called_once_with("handle", game)
        game.get_legal_moves.assert_not_called()
        handler.handle_game_player_turn_event.assert_not_called()
        handler.manager.lookup_player.assert_called_once_with(handle="handle")

    def test_handle_game_next_turn_event_human_not_playing(self):
        for player_state in [player_state for player_state in PlayerState if player_state != PlayerState.PLAYING]:
            player = MagicMock(player_state=player_state)
            game = MagicMock()
            game.get_next_turn.return_value = ("handle", PlayerType.HUMAN)
            game.is_playing.return_value = True
            handler = EventHandler(MagicMock())
            handler.manager.lookup_player = MagicMock(return_value=player)
            handler.handle_game_programmatic_move_event = MagicMock()
            handler.handle_game_player_turn_event = MagicMock()
            handler.handle_game_next_turn_event(game)
            handler.handle_game_programmatic_move_event.assert_called_once_with("handle", game)
            game.get_legal_moves.assert_not_called()
            handler.handle_game_player_turn_event.assert_not_called()
            handler.manager.lookup_player.assert_called_once_with(handle="handle")

    def test_handle_game_next_turn_event_human_playing(self):
        move = MagicMock()
        player = MagicMock(player_state=PlayerState.PLAYING)
        game = MagicMock()
        game.get_next_turn.return_value = ("handle", PlayerType.HUMAN)
        game.is_playing.return_value = True
        game.get_legal_moves.return_value = [move]
        handler = EventHandler(MagicMock())
        handler.manager.lookup_player = MagicMock(return_value=player)
        handler.handle_game_programmatic_move_event = MagicMock()
        handler.handle_game_player_turn_event = MagicMock()
        handler.handle_game_next_turn_event(game)
        handler.handle_game_programmatic_move_event.assert_not_called()
        game.get_legal_moves.assert_called_once_with("handle")
        handler.handle_game_player_turn_event.assert_called_once_with(player, [move])
        handler.manager.lookup_player.assert_called_once_with(handle="handle")

    def test_handle_game_player_change_event(self):
        game_player = MagicMock()
        player = MagicMock()
        game = MagicMock(game_id="game")
        game.get_game_players.return_value = [game_player]
        context = GamePlayerChangeContext(game_id="game", comment="comment", players=[game_player])
        message = Message(MessageType.GAME_PLAYER_CHANGE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.handle_game_player_change_event(game, "comment")
        handler.queue.message.assert_called_once_with(message, players=[player])

    @patch("apologiesserver.event.GameStateChangeContext")
    def test_handle_game_state_change_event_specific_player(self, game_state_change_context):
        history = MagicMock()
        context = GameStateChangeContext("game", player=None, opponents=None, recent_history=[history])
        player = MagicMock(handle="handle")
        view = MagicMock()
        game = MagicMock(game_id="game")
        game.get_player_view.return_value = view
        game.get_recent_history.return_value = [history]
        game_state_change_context.for_context.return_value = context
        message = Message(MessageType.GAME_STATE_CHANGE, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock()
        handler.handle_game_state_change_event(game, player)
        game.mark_active.assert_called_once()
        game.get_player_view.assert_called_with("handle")
        game.get_recent_history.assert_called_once_with(10)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.lookup_game_players.assert_not_called()

    def test_handle_game_state_change_event_game_not_playing(self):
        game = MagicMock(game_id="game")
        game.is_playing.return_value = False
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.manager.lookup_game_players = MagicMock()
        handler.handle_game_state_change_event(game)
        game.mark_active.assert_called_once()
        game.get_player_view.assert_not_called()
        handler.queue.message.assert_not_called()
        handler.manager.lookup_game_players.assert_not_called()

    @patch("apologiesserver.event.GameStateChangeContext")
    def test_handle_game_state_change_event_game_players(self, game_state_change_context):
        game = MagicMock(game_id="game", name="name", mode=GameMode.ADULT, advertiser_handle="advertiser")
        history = MagicMock()
        context = GameStateChangeContext("game", player=None, opponents=None, recent_history=[history])
        player = MagicMock(handle="handle")
        view = MagicMock()
        game = MagicMock(game_id="game")
        game.is_playing.return_value = True
        game.get_player_view.return_value = view
        game.get_recent_history.return_value = [history]
        game_state_change_context.for_context.return_value = context
        message = Message(MessageType.GAME_STATE_CHANGE, context=context)
        handler = EventHandler(MagicMock())
        handler.manager.lookup_game_players = MagicMock(return_value=[player])
        handler.queue.message = MagicMock()
        handler.handle_game_state_change_event(game)
        game.mark_active.assert_called_once()
        game.get_player_view.assert_called_with("handle")
        game.get_recent_history.assert_called_once_with(10)
        handler.queue.message.assert_called_once_with(message, players=[player])
        handler.manager.lookup_game_players.assert_called_once_with(game)

    @patch("apologiesserver.event.GamePlayerTurnContext")
    def test_handle_game_player_turn_event(self, game_player_turn_context):
        context = GamePlayerTurnContext("handle", "game", None, None)
        player = MagicMock(handle="handle", game_id="game")
        moves = [MagicMock()]
        game_player_turn_context.for_moves.return_value = context
        message = Message(MessageType.GAME_PLAYER_TURN, context=context)
        handler = EventHandler(MagicMock())
        handler.queue.message = MagicMock()
        handler.handle_game_player_turn_event(player, moves)
        handler.queue.message.assert_called_once_with(message, players=[player])



# vim: set ft=python ts=4 sw=4 expandtab:

import os
from unittest.mock import AsyncMock

import pytest

from apologiesserver.interface import Message, MessageType
from apologiesserver.util import close, homedir, mask, receive, send, setup_logging
from tests.conftest import coroutine_mock


class TestUtil:
    """
    Unit tests for utilities.
    """

    def test_homedir(self):
        assert homedir() == os.path.expanduser("~")  # different way to get same value

    @pytest.mark.asyncio
    async def test_close(self):
        websocket = AsyncMock()
        websocket.close = coroutine_mock()
        await close(websocket)
        websocket.close.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_send_empty(self):
        message = ""
        websocket = AsyncMock()
        websocket.send = coroutine_mock()
        await send(websocket, message)
        websocket.send.assert_not_awaited()

    @pytest.mark.asyncio
    async def test_send_string(self):
        message = "json"
        websocket = AsyncMock()
        websocket.send = coroutine_mock()
        await send(websocket, message)
        websocket.send.assert_awaited_once_with(message)

    @pytest.mark.asyncio
    async def test_send_message(self):
        message = Message(MessageType.WEBSOCKET_IDLE)
        websocket = AsyncMock()
        websocket.send = coroutine_mock()
        await send(websocket, message)
        websocket.send.assert_awaited_once_with(message.to_json())

    @pytest.mark.asyncio
    async def test_receive_no_timeout(self):
        message = Message(MessageType.WEBSOCKET_INACTIVE)
        data = message.to_json()
        websocket = AsyncMock()
        websocket.recv = coroutine_mock()
        websocket.recv.return_value = data
        assert await receive(websocket, timeout_sec=None) == message

    @pytest.mark.asyncio
    async def test_receive_with_timeout(self):
        message = Message(MessageType.WEBSOCKET_INACTIVE)
        data = message.to_json()
        websocket = AsyncMock()
        websocket.recv = coroutine_mock()
        websocket.recv.return_value = data
        assert await receive(websocket, timeout_sec=3) == message

    def test_mask(self):
        assert mask(None) == ""
        assert mask("") == ""
        assert mask(b"") == ""
        assert mask("hello") == "hello"
        assert mask(b"hello") == "hello"

        assert (
            mask(
                """
        {
          "player_id": null,
        }
        """
            )
            == """
        {
          "player_id": null,
        }
        """
        )

        assert (
            mask(
                """
        {
          "player_id": "",
        }
        """
            )
            == """
        {
          "player_id": "",
        }
        """
        )

        assert (
            mask(
                """
        {
          "player_id": "id",
        }
        """
            )
            == """
        {
          "player_id": "<masked>",
        }
        """
        )

        assert (
            mask(
                """
            {
                 "player_id"    :     "id"  ,
            }
            """
            )
            == """
            {
                 "player_id": "<masked>"  ,
            }
            """
        )

    def test_setup_logging(self):
        setup_logging(quiet=True, verbose=False, debug=False)  # just confirm that it runs
        setup_logging(quiet=False, verbose=True, debug=False)  # just confirm that it runs
        setup_logging(quiet=False, verbose=False, debug=True)  # just confirm that it runs



# vim: set ft=python ts=4 sw=4 expandtab:

"""Scripts that are installed by Poetry in the published package."""

from apologiesserver.cli import cli


def server() -> None:
    """Target for Poetry's script configuration."""
    cli("run_server")



# vim: set ft=python ts=4 sw=4 expandtab:
# ruff: noqa: T201


import pytest
from apologies.game import Card, CardType, GameMode, History, Pawn, Player, PlayerColor, PlayerView, Position
from apologies.rules import Action, ActionType, Move

from apologiesserver.interface import (
    ActivityState,
    AdvertisedGame,
    AdvertiseGameContext,
    AvailableGamesContext,
    CancelledReason,
    ConnectionState,
    ExecuteMoveContext,
    FailureReason,
    GameAdvertisedContext,
    GameCancelledContext,
    GameCompletedContext,
    GameIdleContext,
    GameInactiveContext,
    GameInvitationContext,
    GameJoinedContext,
    GamePlayer,
    GamePlayerChangeContext,
    GamePlayerQuitContext,
    GamePlayerTurnContext,
    GameStartedContext,
    GameStateChangeContext,
    GameStateHistory,
    GameStatePawn,
    JoinGameContext,
    Message,
    MessageType,
    PlayerIdleContext,
    PlayerInactiveContext,
    PlayerMessageReceivedContext,
    PlayerRegisteredContext,
    PlayerState,
    PlayerType,
    ProcessingError,
    RegisteredPlayer,
    RegisteredPlayersContext,
    RegisterPlayerContext,
    RequestFailedContext,
    ReregisterPlayerContext,
    SendMessageContext,
    Visibility,
)
from tests.conftest import to_date


def roundtrip(message: Message) -> None:
    """Round-trip a message to JSON and back, and confirm that it is equivalent."""
    data = message.to_json()
    copy = Message.for_json(data)
    assert message is not copy
    assert message == copy


def create_view() -> PlayerView:
    """Create a realistic-looking player view for testing."""
    player = Player(PlayerColor.RED)
    player.turns = 16
    player.hand = [Card("card1", CardType.CARD_APOLOGIES), Card("card2", CardType.CARD_1)]
    player.pawns[0].position.move_to_square(32)
    player.pawns[1].position.move_to_safe(3)
    player.pawns[2].position.move_to_square(45)

    opponent = Player(PlayerColor.GREEN)
    opponent.turns = 15
    opponent.hand = [Card("card3", CardType.CARD_5), Card("card4", CardType.CARD_11)]
    opponent.pawns[1].position.move_to_home()
    opponent.pawns[2].position.move_to_safe(4)
    opponent.pawns[3].position.move_to_square(19)

    opponents = {opponent.color: opponent}

    return PlayerView(player, opponents)


def create_moves_simple() -> list[Move]:
    """Create a simple move for testing."""
    pawn = Pawn(color=PlayerColor.RED, index=1, name="V", position=Position().move_to_safe(3))
    position = Position().move_to_square(10)

    move = Move(
        id="a9fff13fbe5e46feaeda87382bf4c3b8",
        card=Card("card1", CardType.CARD_APOLOGIES),
        actions=[Action(ActionType.MOVE_TO_POSITION, pawn, position)],
        side_effects=[],
    )

    return [move]


def create_moves_complex() -> list[Move]:
    """Create a complex move for testing."""
    pawn1 = Pawn(color=PlayerColor.RED, index=1, name="V", position=Position().move_to_safe(3))
    position1 = Position().move_to_square(10)

    pawn2 = Pawn(color=PlayerColor.YELLOW, index=3, name="W", position=Position().move_to_square(10))
    position2 = Position().move_to_square(11)

    pawn3 = Pawn(color=PlayerColor.BLUE, index=2, name="X", position=Position().move_to_square(32))

    pawn4 = Pawn(color=PlayerColor.RED, index=3, name="Z", position=Position().move_to_square(17))
    position4 = Position().move_to_square(27)

    move1 = Move(
        id="a9fff13fbe5e46feaeda87382bf4c3b8",
        card=Card("card1", CardType.CARD_APOLOGIES),
        actions=[Action(ActionType.MOVE_TO_POSITION, pawn1, position1), Action(ActionType.MOVE_TO_POSITION, pawn2, position2)],
        side_effects=[Action(ActionType.MOVE_TO_START, pawn3)],
    )

    move2 = Move(
        id="d05f9b511b6e439aa18c8b70cbbcc5d3",
        card=Card("card2", CardType.CARD_10),
        actions=[Action(ActionType.MOVE_TO_POSITION, pawn4, position4)],
        side_effects=[],
    )

    return [move1, move2]


class TestProcessingError:
    def test_no_comment(self) -> None:
        error = ProcessingError(FailureReason.INVALID_PLAYER)
        assert error.reason == FailureReason.INVALID_PLAYER
        assert error.comment is None
        assert "%s" % error == FailureReason.INVALID_PLAYER.value

    def test_comment(self) -> None:
        error = ProcessingError(FailureReason.INVALID_PLAYER, "comment")
        assert error.reason == FailureReason.INVALID_PLAYER
        assert error.comment == "comment"
        assert "%s" % error == "comment"


class TestGameStateChangeContext:
    def test_for_context(self) -> None:
        action = "action"
        color = PlayerColor.RED
        card = CardType.CARD_APOLOGIES
        timestamp = to_date("2020-05-14T13:53:35,334")
        view = create_view()
        history = [History(action, color, card, timestamp)]
        context = GameStateChangeContext.for_context("game", view, history)
        assert context.game_id == "game"
        assert context.recent_history == [GameStateHistory(action, color, card, timestamp)]

        player = context.player
        assert player.color == PlayerColor.RED
        assert player.turns == 16
        assert player.hand == [CardType.CARD_APOLOGIES, CardType.CARD_1]
        assert player.pawns[0] == GameStatePawn(color=PlayerColor.RED, id="0", start=False, home=False, safe=None, square=32)
        assert player.pawns[1] == GameStatePawn(color=PlayerColor.RED, id="1", start=False, home=False, safe=3, square=None)
        assert player.pawns[2] == GameStatePawn(color=PlayerColor.RED, id="2", start=False, home=False, safe=None, square=45)
        assert player.pawns[3] == GameStatePawn(color=PlayerColor.RED, id="3", start=True, home=False, safe=None, square=None)

        assert len(context.opponents) == 1
        opponent = context.opponents[0]
        assert opponent.color == PlayerColor.GREEN
        assert opponent.turns == 15
        assert opponent.hand == [CardType.CARD_5, CardType.CARD_11]
        assert opponent.pawns[0] == GameStatePawn(color=PlayerColor.GREEN, id="0", start=True, home=False, safe=None, square=None)
        assert opponent.pawns[1] == GameStatePawn(color=PlayerColor.GREEN, id="1", start=False, home=True, safe=None, square=None)
        assert opponent.pawns[2] == GameStatePawn(color=PlayerColor.GREEN, id="2", start=False, home=False, safe=4, square=None)
        assert opponent.pawns[3] == GameStatePawn(color=PlayerColor.GREEN, id="3", start=False, home=False, safe=None, square=19)

        message = Message(MessageType.GAME_STATE_CHANGE, context=context)
        print(message.to_json())


class TestGamePlayerTurnContext:
    def test_for_moves_simple(self) -> None:
        moves = create_moves_simple()
        context = GamePlayerTurnContext.for_moves("handle", "game", moves)
        assert context.handle == "handle"
        assert context.game_id == "game"

        assert context.drawn_card == CardType.CARD_APOLOGIES  # because there is only one card among the moves
        assert len(context.moves) == 1

        move = context.moves["a9fff13fbe5e46feaeda87382bf4c3b8"]
        assert move.move_id == "a9fff13fbe5e46feaeda87382bf4c3b8"
        assert move.card == CardType.CARD_APOLOGIES
        assert len(move.actions) == 1
        assert len(move.side_effects) == 0
        assert move.actions[0].start == GameStatePawn(PlayerColor.RED, id="1", start=False, home=False, safe=3, square=None)
        assert move.actions[0].end == GameStatePawn(PlayerColor.RED, id="1", start=False, home=False, safe=None, square=10)

    def test_for_moves_complex(self) -> None:
        moves = create_moves_complex()
        context = GamePlayerTurnContext.for_moves("handle", "game", moves)
        assert context.handle == "handle"
        assert context.game_id == "game"

        assert context.drawn_card is None  # because there is more than one card among the legal moves
        assert len(context.moves) == 2

        move = context.moves["a9fff13fbe5e46feaeda87382bf4c3b8"]
        assert move.move_id == "a9fff13fbe5e46feaeda87382bf4c3b8"
        assert move.card == CardType.CARD_APOLOGIES
        assert len(move.actions) == 2
        assert len(move.side_effects) == 1
        assert move.actions[0].start == GameStatePawn(PlayerColor.RED, id="1", start=False, home=False, safe=3, square=None)
        assert move.actions[0].end == GameStatePawn(PlayerColor.RED, id="1", start=False, home=False, safe=None, square=10)
        assert move.actions[1].start == GameStatePawn(PlayerColor.YELLOW, id="3", start=False, home=False, safe=None, square=10)
        assert move.actions[1].end == GameStatePawn(PlayerColor.YELLOW, id="3", start=False, home=False, safe=None, square=11)
        assert move.side_effects[0].start == GameStatePawn(PlayerColor.BLUE, id="2", start=False, home=False, safe=None, square=32)
        assert move.side_effects[0].end == GameStatePawn(PlayerColor.BLUE, id="2", start=True, home=False, safe=None, square=None)

        move = context.moves["d05f9b511b6e439aa18c8b70cbbcc5d3"]
        assert move.move_id == "d05f9b511b6e439aa18c8b70cbbcc5d3"
        assert move.card == CardType.CARD_10
        assert len(move.actions) == 1
        assert len(move.side_effects) == 0
        assert move.actions[0].start == GameStatePawn(PlayerColor.RED, id="3", start=False, home=False, safe=None, square=17)
        assert move.actions[0].end == GameStatePawn(PlayerColor.RED, id="3", start=False, home=False, safe=None, square=27)


class TestGeneral:
    """General test cases for the public interface functionality."""

    def test_message_invalid_type(self) -> None:
        with pytest.raises(ValueError, match=r"'message' must be a MessageType"):
            Message(None, "id", "Hello")  # type: ignore
        with pytest.raises(ValueError, match=r"'message' must be a MessageType"):
            Message("", "id", "Hello")  # type: ignore
        with pytest.raises(ValueError, match=r"'message' must be a MessageType"):
            Message(PlayerType.HUMAN, "id", "Hello")  # type: ignore

    def test_message_invalid_player_id(self) -> None:
        with pytest.raises(ValueError, match=r"Message type JOIN_GAME requires a player id"):
            Message(MessageType.JOIN_GAME, None, GameJoinedContext("handle", "id", "name", GameMode.STANDARD, "advertiser"))
        with pytest.raises(ValueError, match=r"Message type REGISTER_PLAYER does not allow a player id"):
            Message(MessageType.REGISTER_PLAYER, "id", "Hello")
        with pytest.raises(ValueError, match=r"Message type REQUEST_FAILED does not allow a player id"):
            Message(MessageType.REQUEST_FAILED, "id", RequestFailedContext(FailureReason.INTERNAL_ERROR, "comment"))

    def test_message_invalid_context(self) -> None:
        with pytest.raises(ValueError, match=r"Message type REGISTER_PLAYER requires a context"):
            Message(MessageType.REGISTER_PLAYER, None, None)
        with pytest.raises(ValueError, match=r"Message type JOIN_GAME requires a context"):
            Message(MessageType.JOIN_GAME, "id", None)
        with pytest.raises(ValueError, match=r"Message type GAME_JOINED does not support this context"):
            Message(MessageType.GAME_JOINED, None, "Hello")
        with pytest.raises(ValueError, match=r"Message type WEBSOCKET_INACTIVE does not allow a context"):
            Message(MessageType.WEBSOCKET_INACTIVE, None, "Hello")

    def test_from_json_missing_message(self) -> None:
        data = """
        {
          "wrong": "REGISTER_PLAYER",
          "context": {
            "handle": "leela"
          }
        }
        """
        with pytest.raises(ValueError, match=r"Message type is required"):
            Message.for_json(data)

    def test_from_json_extra_player_id(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER",
          "player_id": "id",
          "context": {
            "handle": "leela"
          }
        }
        """
        with pytest.raises(ValueError, match=r"Message type REGISTER_PLAYER does not allow a player id"):
            Message.for_json(data)

    def test_from_json_missing_player_id(self) -> None:
        data = """
        {
          "message": "REREGISTER_PLAYER",
          "context": {
            "handle": "leela"
          }
        }
        """
        with pytest.raises(ValueError, match=r"Message type REREGISTER_PLAYER requires a player id"):
            Message.for_json(data)

    def test_from_json_missing_context(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER"
        }
        """
        with pytest.raises(ValueError, match=r"Message type REGISTER_PLAYER requires a context"):
            Message.for_json(data)

    def test_from_json_extra_context(self) -> None:
        data = """
        {
          "message": "LIST_AVAILABLE_GAMES",
          "player_id": "id",
          "context": {
            "handle": "leela"
          }
        }
        """
        with pytest.raises(ValueError, match=r"Message type LIST_AVAILABLE_GAMES does not allow a context"):
            Message.for_json(data)

    def test_from_json_wrong_context(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "handle": "leela"
          }
        }
        """
        with pytest.raises(ValueError, match=r"Invalid value 'name'"):
            Message.for_json(data)

    def test_from_json_unknown_message(self) -> None:
        data = """
        {
          "message": "BOGUS",
          "player_id": "id",
          "context": {
            "handle": "leela"
          }
        }
        """
        with pytest.raises(ValueError, match=r"Unknown message type: BOGUS"):
            Message.for_json(data)


class TestRequest:
    """Test cases for request messages."""

    def test_register_player_valid_no_player_id(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER",
          "context": {
            "handle": "aB_-567890123456789012345"
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.REGISTER_PLAYER
        assert message.context.handle == "aB_-567890123456789012345"

    def test_register_player_valid_null_player_id(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER",
          "player_id": null,
          "context": {
            "handle": "aB_-567890123456789012345"
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.REGISTER_PLAYER
        assert message.context.handle == "aB_-567890123456789012345"

    def test_register_player_invalid_handle_none(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER",
          "context": {
            "handle": null
          }
        }
        """
        with pytest.raises(ValueError, match=r"'handle' must be a non-empty string"):
            Message.for_json(data)

    def test_register_player_invalid_handle_empty(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER",
          "context": {
            "handle": ""
          }
        }
        """
        with pytest.raises(ValueError, match=r"'handle' must be a non-empty string"):
            Message.for_json(data)

    def test_register_player_invalid_handle_length(self) -> None:
        data = """
        {
          "message": "REGISTER_PLAYER",
          "context": {
            "handle": "12345678901234567890123456"
          }
        }
        """
        with pytest.raises(ValueError, match=r"'handle' must not exceed length 25"):
            Message.for_json(data)

    def test_register_player_invalid_handle_regex(self) -> None:
        for handle in [";", "ab(c)", "handle name"]:  # just a few examples
            data = (
                """
            {
              "message": "REGISTER_PLAYER",
              "context": {
                "handle": "%s"
              }
            }
            """
                % handle
            )
            with pytest.raises(ValueError, match=r"'handle' does not match regex"):
                Message.for_json(data)

    def test_reregister_player_valid(self) -> None:
        data = """
        {
          "message": "REREGISTER_PLAYER",
          "player_id": "id",
          "context": {
            "handle": "leela"
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.REREGISTER_PLAYER
        assert message.context.handle == "leela"

    def test_reregister_player_invalid_handle_none(self) -> None:
        data = """
        {
          "message": "REREGISTER_PLAYER",
          "player_id": "id",
          "context": {
            "handle": null
          }
        }
        """
        with pytest.raises(ValueError, match=r"'handle' must be a non-empty string"):
            Message.for_json(data)

    def test_reregister_player_invalid_handle_empty(self) -> None:
        data = """
        {
          "message": "REREGISTER_PLAYER",
          "player_id": "id",
          "context": {
            "handle": ""
          }
        }
        """
        with pytest.raises(ValueError, match=r"'handle' must be a non-empty string"):
            Message.for_json(data)

    def test_reregister_player_invalid_handle_length(self) -> None:
        data = """
        {
          "message": "REREGISTER_PLAYER",
          "player_id": "id",
          "context": {
            "handle": "12345678901234567890123456"
          }
        }
        """
        with pytest.raises(ValueError, match=r"'handle' must not exceed length 25"):
            Message.for_json(data)

    def test_advertise_game_valid_handles(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PRIVATE",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.ADVERTISE_GAME
        assert message.context.name == "Leela's Game"
        assert message.context.mode == GameMode.STANDARD
        assert message.context.players == 3
        assert message.context.visibility == Visibility.PRIVATE
        assert message.context.invited_handles == ["bender", "hermes"]

    def test_advertise_game_valid_no_handles(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game 45678901234567890123457890",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": []
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.ADVERTISE_GAME
        assert message.context.name == "Leela's Game 45678901234567890123457890"
        assert message.context.mode == GameMode.STANDARD
        assert message.context.players == 3
        assert message.context.visibility == Visibility.PUBLIC
        assert message.context.invited_handles == []

    def test_advertise_game_invalid_name_none(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": null,
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'name' must be a non-empty string"):
            Message.for_json(data)

    def test_advertise_game_invalid_name_empty(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'name' must be a non-empty string"):
            Message.for_json(data)

    def test_advertise_game_invalid_name_too_long(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "12345678901234567890123456789012345678901",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'name' must not exceed length 40"):
            Message.for_json(data)

    def test_advertise_game_invalid_mode_none(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": null,
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'mode' must be one of \[ADULT, STANDARD\]"):
            Message.for_json(data)

    def test_advertise_game_invalid_mode_empty(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'mode' must be one of \[ADULT, STANDARD\]"):
            Message.for_json(data)

    def test_advertise_game_invalid_mode_bad(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "BOGUS",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"Invalid value 'BOGUS'"):
            Message.for_json(data)

    def test_advertise_game_invalid_player_small(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 1,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'players' must be in \[2, 3, 4\] \(got 1\)"):
            Message.for_json(data)

    def test_advertise_game_invalid_player_large(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 5,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'players' must be in \[2, 3, 4\] \(got 5\)"):
            Message.for_json(data)

    def test_advertise_game_invalid_visibility_none(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": null,
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'visibility' must be one of \[PRIVATE, PUBLIC\]"):
            Message.for_json(data)

    def test_advertise_game_invalid_visibility_empty(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'visibility' must be one of \[PRIVATE, PUBLIC\]"):
            Message.for_json(data)

    def test_advertise_game_invalid_visibility_bad(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "BOGUS",
            "invited_handles": [ "bender", "hermes" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"Invalid value 'BOGUS'"):
            Message.for_json(data)

    def test_advertise_game_invalid_handle_none(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": null
          }
        }
        """
        with pytest.raises(ValueError, match=r"Message type ADVERTISE_GAME does not support this context"):
            Message.for_json(data)

    def test_advertise_game_invalid_handle_none_value(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", null ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'invited_handles' elements must be non-empty strings"):
            Message.for_json(data)

    def test_advertise_game_invalid_handle_empty_value(self) -> None:
        data = """
        {
          "message": "ADVERTISE_GAME",
          "player_id": "id",
          "context": {
            "name": "Leela's Game",
            "mode": "STANDARD",
            "players": 3,
            "visibility": "PUBLIC",
            "invited_handles": [ "bender", "" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'invited_handles' elements must be non-empty strings"):
            Message.for_json(data)

    def test_join_game_valid(self) -> None:
        data = """
        {
          "message": "JOIN_GAME",
          "player_id": "id",
          "context": {
            "game_id": "f13b405e-36e5-45f3-a351-e45bf487acfe"
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.JOIN_GAME
        assert message.context.game_id == "f13b405e-36e5-45f3-a351-e45bf487acfe"

    def test_join_game_invalid_game_id_none(self) -> None:
        data = """
        {
          "message": "JOIN_GAME",
          "player_id": "id",
          "context": {
            "game_id": null
          }
        }
        """
        with pytest.raises(ValueError, match=r"'game_id' must be a non-empty string"):
            Message.for_json(data)

    def test_join_game_invalid_game_id_empty(self) -> None:
        data = """
        {
          "message": "JOIN_GAME",
          "player_id": "id",
          "context": {
            "game_id": ""
          }
        }
        """
        with pytest.raises(ValueError, match=r"'game_id' must be a non-empty string"):
            Message.for_json(data)

    def test_execute_move_valid(self) -> None:
        data = """
        {
          "message": "EXECUTE_MOVE",
          "player_id": "id",
          "context": {
            "move_id": "4"
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.EXECUTE_MOVE
        assert message.context.move_id == "4"

    def test_execute_move_invalid_move_id_none(self) -> None:
        data = """
        {
          "message": "EXECUTE_MOVE",
          "player_id": "id",
          "context": {
            "move_id": null
          }
        }
        """
        with pytest.raises(ValueError, match=r"'move_id' must be a non-empty string"):
            Message.for_json(data)

    def test_execute_move_invalid_move_id_empty(self) -> None:
        data = """
        {
          "message": "EXECUTE_MOVE",
          "player_id": "id",
          "context": {
            "move_id": ""
          }
        }
        """
        with pytest.raises(ValueError, match=r"'move_id' must be a non-empty string"):
            Message.for_json(data)

    def test_optimal_move_valid(self) -> None:
        data = """
        {
          "message": "OPTIMAL_MOVE",
          "player_id": "id"
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.OPTIMAL_MOVE

    def test_send_message_valid(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": "Hello!",
            "recipient_handles": [ "hermes", "nibbler" ]
          }
        }
        """
        message = Message.for_json(data)
        assert message.message == MessageType.SEND_MESSAGE
        assert message.context.message == "Hello!"
        assert message.context.recipient_handles == ["hermes", "nibbler"]

    def test_send_message_invalid_message_none(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": null,
            "recipient_handles": [ "hermes", "nibbler" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'message' must be a non-empty string"):
            Message.for_json(data)

    def test_send_message_invalid_message_empty(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": "",
            "recipient_handles": [ "hermes", "nibbler" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'message' must be a non-empty string"):
            Message.for_json(data)

    def test_send_message_invalid_recipients_none(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": "Hello!",
            "recipient_handles": null
          }
        }
        """
        with pytest.raises(ValueError, match=r"Message type SEND_MESSAGE does not support this context"):
            Message.for_json(data)

    def test_send_message_invalid_recipients_empty(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": "Hello!",
            "recipient_handles": []
          }
        }
        """
        with pytest.raises(ValueError, match=r"'recipient_handles' may not be empty"):
            Message.for_json(data)

    def test_send_message_invalid_recipients_none_value(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": "Hello!",
            "recipient_handles": [ "hermes", null ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'recipient_handles' elements must be non-empty strings"):
            Message.for_json(data)

    def test_send_message_invalid_recipients_empty_value(self) -> None:
        data = """
        {
          "message": "SEND_MESSAGE",
          "player_id": "id",
          "context": {
            "message": "Hello!",
            "recipient_handles": [ "hermes", "" ]
          }
        }
        """
        with pytest.raises(ValueError, match=r"'recipient_handles' elements must be non-empty strings"):
            Message.for_json(data)

    def test_register_player_roundtrip(self) -> None:
        context = RegisterPlayerContext(handle="leela")
        message = Message(MessageType.REGISTER_PLAYER, context=context)
        roundtrip(message)

    def test_reregister_player_roundtrip(self) -> None:
        context = ReregisterPlayerContext(handle="leela")
        message = Message(MessageType.REREGISTER_PLAYER, player_id="id", context=context)
        roundtrip(message)

    def test_unregister_player_roundtrip(self) -> None:
        message = Message(MessageType.UNREGISTER_PLAYER, player_id="id")
        roundtrip(message)

    def test_list_players_roundtrip(self) -> None:
        message = Message(MessageType.LIST_PLAYERS, player_id="id")
        roundtrip(message)

    def test_advertise_game_roundtrip(self) -> None:
        context = AdvertiseGameContext("Leela's Game", GameMode.STANDARD, 3, Visibility.PRIVATE, ["fry", "bender"])
        message = Message(MessageType.ADVERTISE_GAME, player_id="id", context=context)
        roundtrip(message)

    def test_list_available_games_roundtrip(self) -> None:
        message = Message(MessageType.LIST_AVAILABLE_GAMES, player_id="id")
        roundtrip(message)

    def test_join_game_roundtrip(self) -> None:
        context = JoinGameContext("game")
        message = Message(MessageType.JOIN_GAME, player_id="id", context=context)
        roundtrip(message)

    def test_quit_game_roundtrip(self) -> None:
        message = Message(MessageType.QUIT_GAME, player_id="id")
        roundtrip(message)

    def test_start_game_roundtrip(self) -> None:
        message = Message(MessageType.START_GAME, player_id="id")
        roundtrip(message)

    def test_cancel_game_roundtrip(self) -> None:
        message = Message(MessageType.CANCEL_GAME, player_id="id")
        roundtrip(message)

    def test_execute_move_roundtrip(self) -> None:
        context = ExecuteMoveContext("move")
        message = Message(MessageType.EXECUTE_MOVE, player_id="id", context=context)
        roundtrip(message)

    def test_optimal_move_roundtrip(self) -> None:
        message = Message(MessageType.OPTIMAL_MOVE, player_id="id")
        roundtrip(message)

    def test_retrieve_game_state_roundtrip(self) -> None:
        message = Message(MessageType.RETRIEVE_GAME_STATE, player_id="id")
        roundtrip(message)

    def test_send_message_roundtrip(self) -> None:
        context = SendMessageContext("Hello", ["fry", "bender"])
        message = Message(MessageType.SEND_MESSAGE, player_id="id", context=context)
        roundtrip(message)


class TestEvent:
    def test_request_failed_roundtrip(self) -> None:
        context = RequestFailedContext(FailureReason.INTERNAL_ERROR, "it didn't work", "handle")
        message = Message(MessageType.REQUEST_FAILED, context=context)
        roundtrip(message)

    def test_websocket_idle_roundtrip(self) -> None:
        message = Message(MessageType.WEBSOCKET_IDLE)
        roundtrip(message)

    def test_websocket_inactive_roundtrip(self) -> None:
        message = Message(MessageType.WEBSOCKET_INACTIVE)
        roundtrip(message)

    def test_registered_players_roundtrip(self) -> None:
        player = RegisteredPlayer(
            "handle",
            to_date("2020-04-27T09:02:14,334"),
            to_date("2020-04-27T13:19:23,992"),
            ConnectionState.CONNECTED,
            ActivityState.ACTIVE,
            PlayerState.JOINED,
            "game",
        )
        context = RegisteredPlayersContext(players=[player])
        message = Message(MessageType.REGISTERED_PLAYERS, context=context)
        roundtrip(message)

    def test_available_games_roundtrip(self) -> None:
        game = AdvertisedGame("game", "name", GameMode.STANDARD, "leela", 3, 2, Visibility.PUBLIC, ["fry", "bender"])
        context = AvailableGamesContext(games=[game])
        message = Message(MessageType.AVAILABLE_GAMES, context=context)
        roundtrip(message)

    def test_player_registered_roundtrip(self) -> None:
        context = PlayerRegisteredContext("handle")
        message = Message(MessageType.PLAYER_REGISTERED, player_id="player", context=context)
        roundtrip(message)

    def test_player_idle_roundtrip(self) -> None:
        context = PlayerIdleContext("handle")
        message = Message(MessageType.PLAYER_IDLE, context=context)
        roundtrip(message)

    def test_player_inactive_roundtrip(self) -> None:
        context = PlayerInactiveContext("handle")
        message = Message(MessageType.PLAYER_INACTIVE, context=context)
        roundtrip(message)

    def test_player_message_received_roundtrip(self) -> None:
        context = PlayerMessageReceivedContext("leela", ["hermes", "bender"], "Hello")
        message = Message(MessageType.PLAYER_MESSAGE_RECEIVED, context=context)
        roundtrip(message)

    def test_game_advertised_roundtrip(self) -> None:
        game = AdvertisedGame("game", "name", GameMode.STANDARD, "leela", 3, 2, Visibility.PUBLIC, ["fry", "bender"])
        context = GameAdvertisedContext(game)
        message = Message(MessageType.GAME_ADVERTISED, context=context)
        roundtrip(message)

    def test_game_invitation_roundtrip(self) -> None:
        game = AdvertisedGame("game", "name", GameMode.STANDARD, "leela", 3, 2, Visibility.PUBLIC, ["fry", "bender"])
        context = GameInvitationContext(game)
        message = Message(MessageType.GAME_INVITATION, context=context)
        roundtrip(message)

    def test_game_joined_roundtrip(self) -> None:
        context = GameJoinedContext("handle", "game", "name", GameMode.ADULT, "advertiser")
        message = Message(MessageType.GAME_JOINED, context=context)
        roundtrip(message)

    def test_game_started_roundtrip(self) -> None:
        context = GameStartedContext("game")
        message = Message(MessageType.GAME_STARTED, context=context)
        roundtrip(message)

    def test_game_cancelled_roundtrip(self) -> None:
        context = GameCancelledContext("game", CancelledReason.CANCELLED, "YELLOW player (nibbler) quit")
        message = Message(MessageType.GAME_CANCELLED, context=context)
        roundtrip(message)

    def test_game_completed_roundtrip(self) -> None:
        context = GameCompletedContext("game", "nibbler", "Player nibbler won after 46 turns")
        message = Message(MessageType.GAME_COMPLETED, context=context)
        roundtrip(message)

    def test_game_idle_roundtrip(self) -> None:
        context = GameIdleContext("game")
        message = Message(MessageType.GAME_IDLE, context=context)
        roundtrip(message)

    def test_game_inactive_roundtrip(self) -> None:
        context = GameInactiveContext("game")
        message = Message(MessageType.GAME_INACTIVE, context=context)
        roundtrip(message)

    def test_game_player_quit_roundtrip(self) -> None:
        context = GamePlayerQuitContext("handle", "game")
        message = Message(MessageType.GAME_PLAYER_QUIT, context=context)
        roundtrip(message)

    def test_game_player_change_roundtrip(self) -> None:
        red = GamePlayer("leela", PlayerColor.RED, PlayerType.HUMAN, PlayerState.QUIT)
        yellow = GamePlayer("Legolas", PlayerColor.YELLOW, PlayerType.PROGRAMMATIC, PlayerState.PLAYING)
        players = [red, yellow]
        context = GamePlayerChangeContext("game", "YELLOW player (leela) quit", players)
        message = Message(MessageType.GAME_PLAYER_CHANGE, context=context)
        roundtrip(message)

    def test_game_state_change_roundtrip(self) -> None:
        view = create_view()
        recent_history = [History("action", PlayerColor.RED, CardType.CARD_12, to_date("2020-05-14T13:53:35,334"))]
        context = GameStateChangeContext.for_context("game", view, recent_history)
        message = Message(MessageType.GAME_STATE_CHANGE, context=context)
        roundtrip(message)

    def test_game_player_turn_roundtrip(self) -> None:
        moves = create_moves_complex()
        context = GamePlayerTurnContext.for_moves("handle", "game", moves)
        message = Message(MessageType.GAME_PLAYER_TURN, context=context)
        roundtrip(message)



# vim: set ft=python ts=4 sw=4 expandtab:

import os
import pathlib

import pytest

from apologiesserver.config import DEFAULT_CONFIG_PATH, SystemConfig, config, load_config

FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures/test_config")


def assert_defaults(c: SystemConfig) -> None:
    """Check that a SystemConfig object has defaults filled in."""
    assert c.logfile_path is None
    assert c.server_host == "localhost"
    assert c.server_port == 8080
    assert c.close_timeout_sec == 10
    assert c.websocket_limit == 1000
    assert c.total_game_limit == 1000
    assert c.in_progress_game_limit == 25
    assert c.registered_player_limit == 100
    assert c.websocket_idle_thresh_min == 2
    assert c.websocket_inactive_thresh_min == 5
    assert c.player_idle_thresh_min == 15
    assert c.player_inactive_thresh_min == 30
    assert c.game_idle_thresh_min == 10
    assert c.game_inactive_thresh_min == 20
    assert c.game_retention_thresh_min == 2880
    assert c.idle_websocket_check_period_sec == 120
    assert c.idle_websocket_check_delay_sec == 300
    assert c.idle_player_check_period_sec == 120
    assert c.idle_player_check_delay_sec == 300
    assert c.idle_game_check_period_sec == 120
    assert c.idle_game_check_delay_sec == 300
    assert c.obsolete_game_check_period_sec == 300
    assert c.obsolete_game_check_delay_sec == 300


class TestSystemConfig:
    """
    Unit tests for the SystemConfig class.
    """

    def test_defaults(self):
        c = SystemConfig()
        assert_defaults(c)

    def test_to_json(self):
        c = SystemConfig()
        assert c.to_json() is not None


class TestFunctions:
    """
    Unit tests for the config-related functions.
    """

    def test_default_config_path(self):
        assert os.path.join(os.path.expanduser("~"), ".apologiesrc") == DEFAULT_CONFIG_PATH  # different way to get same value

    def test_load_config_defaults(self):
        if not pathlib.Path(DEFAULT_CONFIG_PATH).exists():
            # only bother to test this if the file does not exist; we can't trust what's in there
            load_config()
            assert_defaults(config())

    def test_load_config_path_missing(self):
        config_path = os.path.join(FIXTURE_DIR, "missing.rc")
        assert not pathlib.Path(config_path).exists()  # if it exists, the test isn't valid
        with pytest.raises(ValueError):
            load_config(config_path=config_path)

    def test_load_config_path_exists_partial(self):
        config_path = os.path.join(FIXTURE_DIR, "partial.rc")
        load_config(config_path=config_path)
        c = config()
        assert c.logfile_path == "/var/logs/apologies.log"
        assert c.server_host == "example.com"
        assert c.server_port == 8080
        assert c.close_timeout_sec == 10
        assert c.websocket_limit == 1000
        assert c.total_game_limit == 1000
        assert c.in_progress_game_limit == 25
        assert c.registered_player_limit == 100
        assert c.websocket_idle_thresh_min == 2
        assert c.websocket_inactive_thresh_min == 5
        assert c.player_idle_thresh_min == 15
        assert c.player_inactive_thresh_min == 30
        assert c.game_idle_thresh_min == 600  # taken from partial.rc
        assert c.game_inactive_thresh_min == 700  # taken from partial.rc
        assert c.game_retention_thresh_min == 2880
        assert c.idle_websocket_check_period_sec == 120
        assert c.idle_websocket_check_delay_sec == 300
        assert c.idle_player_check_period_sec == 120
        assert c.idle_player_check_delay_sec == 300
        assert c.idle_game_check_period_sec == 120
        assert c.idle_game_check_delay_sec == 300
        assert c.obsolete_game_check_period_sec == 300
        assert c.obsolete_game_check_delay_sec == 1400  # taken from partial.rc

    def test_load_config_path_exists_complete(self):
        config_path = os.path.join(FIXTURE_DIR, "complete.rc")
        load_config(config_path=config_path)
        c = config()
        assert c.logfile_path == "/var/logs/apologies.log"
        assert c.server_host == "example.com"
        assert c.server_port == 10
        assert c.close_timeout_sec == 20
        assert c.websocket_limit == 30
        assert c.total_game_limit == 40
        assert c.in_progress_game_limit == 50
        assert c.registered_player_limit == 60
        assert c.websocket_idle_thresh_min == 70
        assert c.websocket_inactive_thresh_min == 80
        assert c.player_idle_thresh_min == 90
        assert c.player_inactive_thresh_min == 100
        assert c.game_idle_thresh_min == 110
        assert c.game_inactive_thresh_min == 120
        assert c.game_retention_thresh_min == 130
        assert c.idle_websocket_check_period_sec == 140
        assert c.idle_websocket_check_delay_sec == 150
        assert c.idle_player_check_period_sec == 160
        assert c.idle_player_check_delay_sec == 170
        assert c.idle_game_check_period_sec == 180
        assert c.idle_game_check_delay_sec == 190
        assert c.obsolete_game_check_period_sec == 200
        assert c.obsolete_game_check_delay_sec == 210

    def test_load_config_overrides_partial(self):
        config_path = os.path.join(FIXTURE_DIR, "partial.rc")
        overrides = {
            "logfile_path": "~/logs/apologies.log",
            "server_host": "whatever.com",
            "game_idle_thresh_min": 900,
            "obsolete_game_check_period_sec": 1500,
        }
        load_config(config_path=config_path, overrides=overrides)
        c = config()
        assert c.logfile_path == "~/logs/apologies.log"  # taken from overrides
        assert c.server_host == "whatever.com"  # taken from overrides
        assert c.server_port == 8080
        assert c.close_timeout_sec == 10
        assert c.websocket_limit == 1000
        assert c.total_game_limit == 1000
        assert c.in_progress_game_limit == 25
        assert c.registered_player_limit == 100
        assert c.websocket_idle_thresh_min == 2
        assert c.websocket_inactive_thresh_min == 5
        assert c.player_idle_thresh_min == 15
        assert c.player_inactive_thresh_min == 30
        assert c.game_idle_thresh_min == 900  # taken from overrides
        assert c.game_inactive_thresh_min == 700  # taken from partial.rc
        assert c.game_retention_thresh_min == 2880
        assert c.idle_websocket_check_period_sec == 120
        assert c.idle_websocket_check_delay_sec == 300
        assert c.idle_player_check_period_sec == 120
        assert c.idle_player_check_delay_sec == 300
        assert c.idle_game_check_period_sec == 120
        assert c.idle_game_check_delay_sec == 300
        assert c.obsolete_game_check_period_sec == 1500  # taken from overrides
        assert c.obsolete_game_check_delay_sec == 1400  # taken from partial.rc

    def test_load_config_overrides_all(self):
        config_path = os.path.join(FIXTURE_DIR, "complete.rc")
        overrides = {
            "logfile_path": "~/logs/apologies.log",
            "server_host": "whatever.com",
            "server_port": 100,
            "close_timeout_sec": 200,
            "websocket_limit": 300,
            "total_game_limit": 400,
            "in_progress_game_limit": 500,
            "registered_player_limit": 600,
            "websocket_idle_thresh_min": 700,
            "websocket_inactive_thresh_min": 800,
            "player_idle_thresh_min": 900,
            "player_inactive_thresh_min": 1000,
            "game_idle_thresh_min": 1100,
            "game_inactive_thresh_min": 1200,
            "game_retention_thresh_min": 1300,
            "idle_websocket_check_period_sec": 1400,
            "idle_websocket_check_delay_sec": 1500,
            "idle_player_check_period_sec": 1600,
            "idle_player_check_delay_sec": 1700,
            "idle_game_check_period_sec": 1800,
            "idle_game_check_delay_sec": 1900,
            "obsolete_game_check_period_sec": 2000,
            "obsolete_game_check_delay_sec": 2100,
        }
        load_config(config_path=config_path, overrides=overrides)
        c = config()
        assert c.logfile_path == "~/logs/apologies.log"
        assert c.server_host == "whatever.com"
        assert c.server_port == 100
        assert c.close_timeout_sec == 200
        assert c.websocket_limit == 300
        assert c.total_game_limit == 400
        assert c.in_progress_game_limit == 500
        assert c.registered_player_limit == 600
        assert c.websocket_idle_thresh_min == 700
        assert c.websocket_inactive_thresh_min == 800
        assert c.player_idle_thresh_min == 900
        assert c.player_inactive_thresh_min == 1000
        assert c.game_idle_thresh_min == 1100
        assert c.game_inactive_thresh_min == 1200
        assert c.game_retention_thresh_min == 1300
        assert c.idle_websocket_check_period_sec == 1400
        assert c.idle_websocket_check_delay_sec == 1500
        assert c.idle_player_check_period_sec == 1600
        assert c.idle_player_check_delay_sec == 1700
        assert c.idle_game_check_period_sec == 1800
        assert c.idle_game_check_delay_sec == 1900
        assert c.obsolete_game_check_period_sec == 2000
        assert c.obsolete_game_check_delay_sec == 2100



import random
import string
from unittest.mock import AsyncMock, MagicMock

from arrow import Arrow
from arrow import get as arrow_get


def coroutine_mock() -> AsyncMock:
    # This replaces asynctest's CoroutineMock, which is what I originally used when writing
    # this code.  Unfortunately, even when I was writing this back in 2020, asynctest had
    # effectively been abandoned, and I didn't realize that.  It doesn't work with Python
    # 3.11 or later.
    #
    # Source: https://github.com/openwallet-foundation/acapy/blob/main/acapy_agent/tests/mock.py
    # See also: https://github.com/openwallet-foundation/acapy/pull/2566
    return AsyncMock(return_value=MagicMock())


def random_string(length: int = 10) -> str:
    chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
    return "".join([random.choice(chars) for _ in range(length)])  # noqa: S311


def to_date(date: str) -> Arrow:
    return arrow_get(date)


def mock_handler() -> MagicMock:
    """Create a mocked handler that can be locked and used as expected."""
    lock = coroutine_mock()
    handler = MagicMock()
    handler.__enter__.return_value = handler
    handler.execute_tasks = coroutine_mock()
    handler.manager = MagicMock()
    handler.manager.lock = lock
    handler.manager.lock.__aenter__ = lock
    handler.manager.lock.__aexit__ = lock
    return handler



# vim: set ft=python ts=4 sw=4 expandtab:

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from apologiesserver.scheduled import (
    _execute_idle_game_check,
    _execute_idle_player_check,
    _execute_idle_websocket_check,
    _execute_obsolete_game_check,
    _schedule_idle_game_check,
    _schedule_idle_player_check,
    _schedule_idle_websocket_check,
    _schedule_obsolete_game_check,
    scheduled_tasks,
)
from tests.conftest import coroutine_mock, mock_handler


class TestFunctions:
    """
    Test Python functions.
    """

    @patch("apologiesserver.scheduled._schedule_obsolete_game_check")
    @patch("apologiesserver.scheduled._schedule_idle_game_check")
    @patch("apologiesserver.scheduled._schedule_idle_player_check")
    @patch("apologiesserver.scheduled._schedule_idle_websocket_check")
    def test_scheduled_tasks(
        self, _schedule_idle_websocket_check, schedule_idle_player_check, schedule_idle_game_check, schedule_obsolete_game_check
    ):
        assert scheduled_tasks() == [
            _schedule_idle_websocket_check,
            schedule_idle_player_check,
            schedule_idle_game_check,
            schedule_obsolete_game_check,
        ]


class TestCoroutines:
    """
    Test Python coroutines.
    """

    pytestmark = pytest.mark.asyncio

    @patch("apologiesserver.scheduled.EventHandler")
    @patch("apologiesserver.scheduled.manager")
    async def test_execute_idle_websocket_check(self, manager, event_handler):
        handler = mock_handler()
        manager.return_value = handler.manager
        event_handler.return_value = handler
        await _execute_idle_websocket_check()
        event_handler.assert_called_once_with(manager.return_value)
        handler.manager.lock.assert_awaited()
        handler.handle_idle_websocket_check_task.assert_called_once()
        handler.execute_tasks.assert_awaited_once()

    @patch("apologiesserver.scheduled.EventHandler")
    @patch("apologiesserver.scheduled.manager")
    async def test_execute_idle_player_check(self, manager, event_handler):
        handler = mock_handler()
        manager.return_value = handler.manager
        event_handler.return_value = handler
        await _execute_idle_player_check()
        event_handler.assert_called_once_with(manager.return_value)
        handler.manager.lock.assert_awaited()
        handler.handle_idle_player_check_task.assert_called_once()
        handler.execute_tasks.assert_awaited_once()

    @patch("apologiesserver.scheduled.EventHandler")
    @patch("apologiesserver.scheduled.manager")
    async def test_execute_idle_game_check(self, manager, event_handler):
        handler = mock_handler()
        manager.return_value = handler.manager
        event_handler.return_value = handler
        await _execute_idle_game_check()
        event_handler.assert_called_once_with(manager.return_value)
        handler.manager.lock.assert_awaited()
        handler.handle_idle_game_check_task.assert_called_once()
        handler.execute_tasks.assert_awaited_once()

    @patch("apologiesserver.scheduled.EventHandler")
    @patch("apologiesserver.scheduled.manager")
    async def test_execute_obsolete_game_check(self, manager, event_handler):
        handler = mock_handler()
        manager.return_value = handler.manager
        event_handler.return_value = handler
        await _execute_obsolete_game_check()
        event_handler.assert_called_once_with(manager.return_value)
        handler.manager.lock.assert_awaited()
        handler.handle_obsolete_game_check_task.assert_called_once()
        handler.execute_tasks.assert_awaited_once()

    @patch("apologiesserver.scheduled.config")
    @patch("apologiesserver.scheduled.Periodic", autospec=True)
    async def test_schedule_idle_websocket_check(self, periodic, config):
        p = AsyncMock()
        p.start = coroutine_mock()
        periodic.return_value = p
        config.return_value = MagicMock(idle_websocket_check_period_sec=1, idle_websocket_check_delay_sec=2)
        await _schedule_idle_websocket_check()
        p.start.assert_awaited_once_with(delay=2)
        periodic.assert_called_once_with(1, _execute_idle_websocket_check)

    @patch("apologiesserver.scheduled.config")
    @patch("apologiesserver.scheduled.Periodic", autospec=True)
    async def test_schedule_idle_player_check(self, periodic, config):
        p = AsyncMock()
        p.start = coroutine_mock()
        periodic.return_value = p
        config.return_value = MagicMock(idle_player_check_period_sec=1, idle_player_check_delay_sec=2)
        await _schedule_idle_player_check()
        p.start.assert_awaited_once_with(delay=2)
        periodic.assert_called_once_with(1, _execute_idle_player_check)

    @patch("apologiesserver.scheduled.config")
    @patch("apologiesserver.scheduled.Periodic", autospec=True)
    async def test_schedule_idle_game_check(self, periodic, config):
        p = AsyncMock()
        p.start = coroutine_mock()
        periodic.return_value = p
        config.return_value = MagicMock(idle_game_check_period_sec=1, idle_game_check_delay_sec=2)
        await _schedule_idle_game_check()
        p.start.assert_awaited_once_with(delay=2)
        periodic.assert_called_once_with(1, _execute_idle_game_check)

    @patch("apologiesserver.scheduled.config")
    @patch("apologiesserver.scheduled.Periodic", autospec=True)
    async def test_schedule_obsolete_game_check(self, periodic, config):
        p = AsyncMock()
        p.start = coroutine_mock()
        periodic.return_value = p
        config.return_value = MagicMock(obsolete_game_check_period_sec=1, obsolete_game_check_delay_sec=2)
        await _schedule_obsolete_game_check()
        p.start.assert_awaited_once_with(delay=2)
        periodic.assert_called_once_with(1, _execute_obsolete_game_check)
